from dspsim.framework import Context, Clock, signal, dff
from dspsim.library import AsyncSync8
from dspsim.axis import Axis, AxisTx, AxisRx
from dspsim.fifo import FifoAsync

from dspsim.i2s import I2S, I2SClkGen, I2STx, I2SRx
import numpy as np


# @framework.runner(time_unit=1e-9, time_precision=1e-9)
def test_i2s_loopback():
    with Context(1e-9, 1e-9) as context:
        fs = 48e3
        fs_mclk = fs * 384
        T_mclk = 1 / fs_mclk  # About 54ns

        # Main bus clock.
        bclk = Clock(10e-9)
        # Audio clock.
        mclk = Clock(T_mclk)

        rst = dff(bclk, 1)
        irst = signal()
        # Synchronize the master reset with the i2s clk.
        AsyncSync8(mclk, rst, irst)

        # i2s bus
        i2s = I2S(mclk)
        # If the fpga is generating the i2s clock, use the clkgen component.
        i2s_clk_gen = I2SClkGen.init_bus(mclk, irst, i2s)

        tx_data_bus = Axis(width=I2STx.DW, tid=True)
        tx_fifo_bus = Axis(width=I2STx.DW, tid=True)
        i2s_rx_bus = Axis(width=I2SRx.DW, tid=True)
        rx_fifo_bus = Axis(width=I2SRx.DW, tid=True)

        # Send data to the fifo component. Alternates tid.
        axis_tx = AxisTx(bclk, rst, tx_data_bus, tid_pattern=[1, 0], width=I2STx.DW)

        # Cross clock domains with an async fifo.
        fifo_tx = FifoAsync.init_bus(bclk, rst, tx_data_bus, mclk, irst, tx_fifo_bus)

        # I2S component receives data from fifo, then sends it on the i2s bus.
        i2s_tx = I2STx.init_bus(mclk, irst, tx_fifo_bus, i2s)

        # I2S component receives i2s data, then sends it to a fifo.
        i2s_rx = I2SRx.init_bus(mclk, irst, i2s_rx_bus, i2s)

        # Fifo crosses clock domains back to bus clock.
        fifo_rx = FifoAsync.init_bus(mclk, irst, i2s_rx_bus, bclk, rst, rx_fifo_bus)

        # Receive axis data.
        axis_rx = AxisRx(bclk, rst, rx_fifo_bus)

        print(context.print_info())

        i2s_clk_gen.trace("traces/i2s_clk_gen.vcd")
        i2s_tx.trace("traces/i2s_tx.vcd")
        i2s_rx.trace("traces/i2s_rx.vcd")
        fifo_tx.trace("traces/i2s_fifo_tx.vcd")
        fifo_rx.trace("traces/i2s_fifo_rx.vcd")

        context.elaborate()

        tx_data = np.linspace(1, 42, 47, dtype=np.int32)
        print(tx_data)
        axis_tx.write(tx_data)

        rst.d = 1
        context.advance(500)
        rst.d = 0
        context.advance(100)
        axis_rx.tready = True

        context.advance(1000000)

        rx_data = axis_rx.read()
        print(rx_data)
        # print(rx_tid)
        # assert all(rx_data == tx_data)



"""
The SomeModel.sv component contains every type of parameter and port.
It can be used to test code generation and bus connection.
"""

from dspsim.framework import Context, Clock, signal, dff
from dspsim.library import SomeModel, Foo, X


def test_context_basic():
    with Context(1e-9, 1e-9) as context:
        clk = Clock(10e-9)
        rst = dff(clk, 1)
        x = dff(clk, 42, width=24)
        y = signal(width=24)
        z = signal(width=24)

        c = signal(43, width=18, shape=(SomeModel.NC,))
        d = signal(width=18, shape=(SomeModel.ND, SomeModel.MD))
        e = signal(width=18, shape=(SomeModel.NE, SomeModel.ME))

        print(context.print_info())
        some_model = SomeModel(clk, rst, x, y, c, d, e)

        xmodel = X()
        foo = Foo(clk, rst, x, z)
        print(context.print_info())

        some_model.trace("traces/some_model.vcd")
        foo.trace("traces/foo.vcd")

        context.elaborate()

        context.advance(100)
        print(context.print_info())
        rst.d = 0
        context.advance(100)

        for _ in range(10):
            x.d = x.q + 1
            context.advance(10)



"""
Test the verilator utilities in the dspsim package.
"""

from dspsim.config import Parameter, Port, ModuleConfig
from dspsim.verilator import VModelTree

from pathlib import Path

from dspsim.util import render_template


def test_model_gen():
    """"""
    source = Path("src/dspsim/hdl/SomeModel.sv")

    model = ModuleConfig.from_verilator(source, parameters={}, verilator_args=[])

    model_gen = render_template("model.cpp.jinja", model=model, trace="vcd")
    print(model_gen)



from dspsim.framework import Signal8, signal
from dspsim.axis import Axis

from dspsim._library import I2SClkGen as _I2SClkGen
from dspsim._library import I2STx as _I2STx
from dspsim._library import I2SRx as _I2SRx


class I2S:
    mclk: Signal8
    lrclk: Signal8
    sclk: Signal8
    sd: Signal8

    def __init__(
        self, mclk: Signal8, lrclk: Signal8 = None, sclk: Signal8 = None
    ) -> None:
        # Clocks will be asosciated with a bus.
        self.mclk = mclk
        self.lrclk = lrclk if lrclk else signal()
        self.sclk = sclk if sclk else signal()
        # Create a data signal for this bus.
        self.sd = signal()


class I2SClkGen(_I2SClkGen):
    @classmethod
    def init_bus(cls, clk: Signal8, rst: Signal8, i2s: I2S):
        return cls(clk, rst, i2s.lrclk, i2s.sclk)


class I2STx(_I2STx):
    @classmethod
    def init_bus(cls, clk: Signal8, rst: Signal8, s_axis: Axis, i2s: I2S):
        """"""
        return cls(
            clk,
            rst,
            s_axis.tdata,
            s_axis.tvalid,
            s_axis.tready,
            s_axis.tid,
            i2s.lrclk,
            i2s.sclk,
            i2s.sd,
        )


class I2SRx(_I2SRx):
    @classmethod
    def init_bus(cls, clk: Signal8, rst: Signal8, m_axis: Axis, i2s: I2S):
        """"""
        return cls(
            clk,
            rst,
            m_axis.tdata,
            m_axis.tvalid,
            m_axis.tready,
            m_axis.tid,
            i2s.lrclk,
            i2s.sclk,
            i2s.sd,
        )



"""
Main Command-line app interface "dspsim"
"""

from dataclasses import dataclass, field
from pathlib import Path
import argparse
from typing import Callable
from dspsim.generate import ArgsGenerate


@dataclass
class Args:
    cmake_dir: bool
    include_dir: bool
    hdl_dir: bool
    func: Callable

    @classmethod
    def create_parser(cls):
        """"""
        from . import __version__

        parser = argparse.ArgumentParser("dspsim")
        parser.add_argument(
            "--version",
            action="version",
            version=__version__,
            help="Print the package version.",
        )
        parser.add_argument(
            "--cmake-dir", action="store_true", help="Print the cmake config location."
        )
        parser.add_argument(
            "--include-dir",
            action="store_true",
            help="Print the C++ include directory.",
        )
        parser.add_argument(
            "--hdl-dir", action="store_true", help="Print the HDL directory."
        )
        parser.set_defaults(func=Args)
        subparsers = parser.add_subparsers()
        generate_parser = ArgsGenerate.create_parser(subparsers)

        return parser

    @classmethod
    def parse_args(cls, cli_args: list[str] = None):
        parser = Args.create_parser()
        parsed_args = parser.parse_args(cli_args)
        args: Args = parsed_args.func(**vars(parsed_args))
        return args

    def exec(self):
        """"""
        from .util import cmake_dir, include_dir, hdl_dir

        print(f"Running main_func: {self}")
        if self.cmake_dir:
            print(cmake_dir())
        if self.include_dir:
            print(include_dir())
        if self.hdl_dir:
            print(hdl_dir())


def main(cli_args: list[str] = None):

    args = Args.parse_args(cli_args)
    args.exec()


if __name__ == "__main__":
    exit(main())
